/*******************************************************************************
  File: XdrStream_reader.cpp
  Copyright (c) 1999 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee,
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived 
from
  or uses this Software. You may not do anything however, to prevent the
  continued distribution of this Software under an open source distribution
  model. This includes, but is not limited to, such things as copyrighting
  or claiming authorship of the Software.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
$Header:$
  Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  Implements the basic operations of XDR_Stream. See also 
XDRStream_reader.cpp
  for the implementation of iXDR_Stream, and XDRStream_writer.cpp for the
  implementation of oXDR_Stream and XDR_Stream.

//#----------------------------------------------------------------------------|
Notes:

//*****************************************************************************/

#include "XDRStream.h"

#include "platform.h"

// Standard C++ headers
#include <algorithm>

namespace Util {

//
// iXDR_Stream
//
iXDR_Stream::iXDR_Stream(XDR_Streambuf* sb)
  : inherited(sb)
{
	// empty
}


iXDR_Stream::~iXDR_Stream()
{
	// empty
}


iXDR_Stream&
iXDR_Stream::operator>>(iXDR_Stream& (*pf)(iXDR_Stream&))
{
	pf(*this);
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(std::basic_ios<XDR_Char>& 
(*pf)(std::basic_ios<XDR_Char>&))
{
	pf(*this);
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(std::ios_base& (*pf)(std::ios_base&))
{
	pf(*this);
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(bool& n)
{
	long x;
	*this >> x;
	n = (x == 1) ? true : false;
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(char& n)
{
	long x;
	*this >> x;
	n = static_cast<signed char>(x);
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(signed char& n)
{
	long x;
	*this >> x;
	n = x;
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(unsigned char& n)
{
	unsigned long x;
	*this >> x;
	n = x;
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(short& n)
{
	long x;
	*this >> x;
	n = x;
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(unsigned short& n)
{
	unsigned long x;
	*this >> x;
	n = x;
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(int& n)
{
	long x;
	*this >> x;
	n = x;
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(unsigned int& n)
{
	unsigned long x;
	*this >> x;
	n = x;
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(long& n)
{
	sentry s(*this);
	if (s) {
		XDR_Char c;
		get(c);
#ifdef LITTLE_ENDIAN
		c.swap_bytes();
#endif
		::memcpy(&n, c.data, sizeof(n));
	}
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(unsigned long& n)
{
	sentry s(*this);
	if (s) {
		XDR_Char c;
		get(c);
#ifdef LITTLE_ENDIAN
		c.swap_bytes();
#endif
		::memcpy(&n, c.data, sizeof(n));
	}
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(float& f)
{
	sentry s(*this);
	if (s) {
		XDR_Char c;
		get(c);
#ifdef LITTLE_ENDIAN
		c.swap_bytes();
#endif
		::memcpy(&f, c.data, sizeof(f));
	}
	return *this;
}


iXDR_Stream&
iXDR_Stream::operator>>(double& f)
{
	sentry s(*this);
	if (s) {
		XDR_Char c[2];
		read(c, 2);
#ifdef LITTLE_ENDIAN
		c[0].swap_bytes();
		c[1].swap_bytes();
		std::swap(c[0], c[1]);
#endif
		::memcpy(&f, c, sizeof(f));
	}
	return *this;
}


size_t
iXDR_Stream::get(void* d, size_t l)
{
	unsigned char* data = static_cast<unsigned char*>(d);
	sentry s(*this);
	if (s) {
		while (l > 0) {
			XDR_Char c;
			get(c);
			int k = (l > 4) ? 4 : l;
			for (int i = 0; i < k; ++i) {
				data[i] = c.data[i];
			}
			data += k;
			l -= k;
		}
	}
	return data - static_cast<unsigned char*>(d);
}


size_t
iXDR_Stream::vget(void* d, size_t l)
{
	unsigned long len;
	*this >> len;
	if (len > l)  {} // TBD - error handling
	return get(d, len);
}


size_t
iXDR_Stream::vget(std::vector<unsigned char>& d)
{
	sentry s(*this);
	if (s) {
		d.clear();
		unsigned long l;
		*this >> l;
		while (l > 0) {
			XDR_Char c;
			get(c);
			int k = (l > 4) ? 4 : l;
			for (int i = 0; i < k; ++i) {
				d.push_back(c.data[i]);
			}
			l -= k;
		}
	}
	return d.size();
}


template<typename T>
size_t
iXDR_Stream::get_arr(T* a, size_t l)
{
	T* p = a;
	sentry s(*this);
	if (s) {
		while (l > 0) {
			*this >> *p;
			++p;
			--l;
		}
	}
	return p - a;
}

template<typename T>
size_t
iXDR_Stream::vget_arr(T* a, size_t l)
{
	unsigned long len;
	*this >> len;
	if (len > l) {}  // TBD - error handling
	return get_arr(a, len);
}


template<typename T>
size_t
iXDR_Stream::vget_arr(std::vector<T>& v)
{
	sentry s(*this);
	v.clear();
	if (s) {
		unsigned long l;
		*this >> l;
		while (l > 0) {
			T t;
			*this >> t;
			v.push_back(t);
			--l;
		}
	}
	return v.size();
}


iXDR_Stream::int_type
iXDR_Stream::get()
{
	sentry s(*this);
	if (s) {
		return rdbuf()->sbumpc();
	}
	return traits_type::eof();
}


iXDR_Stream&
iXDR_Stream::get(char_type& c)
{
	int_type t = get();
	c = traits_type::to_char_type(t);
	return *this;
}


iXDR_Stream&
iXDR_Stream::read(char_type* s, std::streamsize n)
{
	sentry guard(*this);
	if (guard) {
		while (n > 0) {
			get(*s);
			++s;
			--n;
		}
	}
	return *this;
}


iXDR_Stream&
operator>>(iXDR_Stream& is, char* s)
{
	return operator>>(is, reinterpret_cast<unsigned char*>(s));
}



iXDR_Stream&
operator>>(iXDR_Stream& is, signed char* s)
{
	return operator>>(is, reinterpret_cast<unsigned char*>(s));
}


iXDR_Stream&
operator>>(iXDR_Stream& is, unsigned char* s)
{
	iXDR_Stream::sentry guard(is);
	if (guard) {
		unsigned long l;
		is >> l;
		if (is.width() > 0 && l > is.width()) {}  // TBD - error
		while (l > 0) {
			XDR_Char c;
			is.get(c);
			int k = (l > 4) ? 4 : l;
			for (int i = 0; i < k; ++i)
				*s++ = c.data[i];
			l += k;
		}
		*s = '\0';
	}
	return is;
}


iXDR_Stream&
operator>>(iXDR_Stream& is, std::string& str)
{
	iXDR_Stream::sentry guard(is);
	if (guard) {
		str.erase();
		unsigned long l;
		is >> l;
		if (is.width() > 0 && l > is.width()) {}  // TBD - error
		while (l > 0) {
			XDR_Char c;
			is.get(c);
			int k = (l > 4) ? 4 : l;
			for (int i = 0; i < k; ++i)
				str += c.data[i];
			l -= k;
		}
	}
	return is;
}


iXDR_Stream::sentry::sentry(iXDR_Stream& is)
  : _ok(is.good())
{
	// empty
}


iXDR_Stream::sentry::~sentry()
{
	// empty
}


/*** Not yet implemented
iXDR_Stream&
iXDR_Stream::operator>>(long double& f);		// later

iXDR_Stream&
iXDR_Stream::operator>>(XDR_Streambuf* sb);		// later

std::streamsize
iXDR_Stream::gcount() const;

iXDR_Stream&
iXDR_Stream::get(XDR_Streambuf& sb, std::streamsize n);		// later

iXDR_Stream&
iXDR_Stream::ignore(std::streamsize n, int_type delim);	// later

iXDR_Stream::int_type
iXDR_Stream::peek();		// later

std::streamsize
iXDR_Stream::readsome(char_type* s, std::streamsize n);		// later

iXDR_Stream&
iXDR_Stream::putback(char_type c);		// later

iXDR_Stream&
iXDR_Stream::unget();					// later

int
iXDR_Stream::sync();					// later

iXDR_Stream::pos_type
iXDR_Stream::tellg();					// later

iXDR_Stream&
iXDR_Stream::seekg(pos_type);			// later

iXDR_Stream&
iXDR_Stream::seekg(off_type, std::ios_base::seekdir);		// later
***/

}  //> namespace Util


