#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <stdarg.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>

#include "comprio.h"
#include "cdc.h"

#define COUNT(x) (sizeof(x)/sizeof((x)[0]))

static struct
{
  unsigned  alloc; /* # of entries alloced in table */
  cdc_t   **tbl;   /* the table                     */
} fd_table;


/*
 * NAME
 *   cdc_reset
 *
 * DESCRIPTION
 *   reset a cdc to its initial value
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
static int cdc_reset(cdc_t *cdc)
{
  memset(&cdc->hdr, 0, sizeof(cdc->hdr));
  
  cdc->hdr.magic     = COMPRIO_MAGIC;
  cdc->hdr.version   = COMPRIO_VERSION;
  return 0;
}

/*
 * NAME
 *   fd_table_entry_alloc
 *
 * DESCRIPTION
 *   get a new table entry
 *
 * PARAMETERS
 *   dst : [out] holds result on success
 *
 * RETURN
 *
 * NOTES
 */
static int fd_table_entry_alloc(int *dst_fd, cdc_t **dst_cdc)
{
  int fd;
  int rc;

  rc = 0;
  for (fd = 0; (fd < fd_table.alloc) && (0 != fd_table.tbl[fd]); fd++)
    ;
  if (fd == fd_table.alloc)
  {
    size_t ct;
    void  *ptr;

    ct = fd_table.alloc;
    if (0 == ct)
    {
      ct = 16;
    } else
    {
      ct *= 2;
    }
    ptr = realloc(fd_table.tbl, ct * sizeof(*fd_table.tbl));
    if (0 == ptr)
    {
      rc = ENOMEM;
    } else
    {
      fd_table.alloc = ct;
      fd_table.tbl   = ptr;
    }
  }
  if (0 == rc)
  {
    cdc_t *cdc;

    cdc = calloc(1, sizeof(*cdc));
    if (0 == cdc)
    {
      rc = ENOMEM;
    } else
    {
      fd_table.tbl[fd]   = cdc;
      cdc_reset(cdc);
      *dst_fd  = fd;
      *dst_cdc = cdc;
    }
  }
  return rc;
}

/*
 * NAME
 *   fd_table_entry_free
 *
 * DESCRIPTION
 *   free a table entry
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
static int fd_table_entry_free(int fd)
{
  cdc_t *cdc = fd_table.tbl[fd];

  free(cdc);
  fd_table.tbl[fd] = 0;
  return 0;
}

/*
 * NAME
 *   fd_table_entry_to_cdc
 *
 * DESCRIPTION
 *   return an open cdc given a handle
 *
 * PARAMETERS
 *   fd  : handle
 *   dst : [out] holds result on success
 *
 * RETURN
 *   0     : no error
 *   EBADF : bad file descriptor
 *
 * NOTES
 */
static int fd_table_entry_to_cdc(int fd, cdc_t **dst)
{
  int rc;

  if ((0 <= fd) && (fd < fd_table.alloc) && (0 != fd_table.tbl[fd]))
  {
    rc = 0;
    *dst = fd_table.tbl[fd];
  } else
  {
    rc = EBADF;
  }
  return rc;
}

static int cdc_cleanup(cdc_t *cdc)
{
  cdc_data_cleanup(cdc);
  cdc_bat_cleanup(cdc);
  cdc_map_cleanup(cdc);
  cdc_blk_cleanup(cdc);
  free(cdc->hdr.buffer);
  cdc->hdr.buffer = 0;
  free(cdc->filename);
  cdc->filename = 0;
  return 0;
}

/*
 * NAME
 *   comprio_open
 *
 * DESCRIPTION
 *   open a file
 *
 * PARAMETERS
 *   env   : environment (0 to use default)
 *   fname : file name
 *   flags : open flags (see open())
 *   [mode]: create mode (only if O_CREAT is set)
 *
 * RETURN
 *    >= 0  : file descriptor
 *     -1   : errno set appropriatly
 *
 * NOTES
 */
int comprio_open(comprio_env_t *env, const char *filename, int flags, ...)
{
  int    fd;
  cdc_t *cdc;

  if (flags & O_APPEND)
  {
    errno = EINVAL;
    fd = -1;
  } else
  {
    int rc;
    int env_local;

    if (0 == env)
    {
      rc = comprio_env_alloc(&env);
      env_local = 1;
    } else
    {
      rc        = 0;
      env_local = 0;
    }
    if (0 == rc)
    {
      rc = fd_table_entry_alloc(&fd, &cdc);
      if (0 == rc)
      {
        int    mode =  0;
        size_t blk_size;
        size_t blk_per_cluster;
        size_t blk_per_map;
        size_t blk_per_bat;

        cdc->env = env;
        blk_size        = env->blk_size;
        blk_per_cluster = env->blk_per_cluster;
        blk_per_map     = env->blk_per_map;
        blk_per_bat     = env->blk_per_bat;
        if (flags & O_CREAT)
        {
          va_list lst;

          va_start(lst, flags);
          mode = va_arg(lst, int);
          va_end(lst);
        }
        if (flags & O_TRUNC)
        {
          /*
           * try to read the existing header to get the file defaults
           */
          cdc->fd = cdc->env->open(filename, O_RDONLY);
          if (cdc->fd >= 0)
          {
            struct stat st;

            rc = cdc->env->stat(cdc->fd, &st);
            if ((0 == rc) && (0 < st.st_size))
            {
              rc = cdc_hdr_read(cdc);
              if (0 == rc)
              {
                blk_size        = cdc->blk.size;
                blk_per_cluster = cdc->data.blocks;
                blk_per_map     = cdc->map.tbl.blocks;
                blk_per_bat     = cdc->bat.tbl.blocks;
                /*
                 * dispose of the header
                 */
                cdc_cleanup(cdc);
              } else if (EINVAL == rc)
              {
                cdc_reset(cdc);
                rc = 0; /* bad file, ignore */
              }
            }
            cdc->env->close(cdc->fd);
          }
        }
        if (0 == rc)
        {
          cdc->filename = strdup(filename);
          if (0 == cdc->filename)
          {
            rc = ENOMEM;
          } else
          {
            cdc->fd = cdc->env->open(filename, flags, mode);
            if (-1 == cdc->fd)
            {
              rc = errno;
            } else
            {
              /*
               * try to read the header
               */
              struct stat st;

              rc = cdc->env->stat(cdc->fd, &st);
              if (0 == (flags & O_CREAT))
              {
                if (0 == rc)
                {
                  rc = cdc_hdr_read(cdc);
                } 
              } else
              {
                /*
                 * create the file
                 */
                cdc->hdr.file_size = 0;

                cdc_blk_init(cdc,  blk_size);
                cdc_data_init(cdc, blk_per_cluster);
                cdc_map_init(cdc,  blk_per_map);
                cdc_bat_init(cdc,  blk_per_bat);
                
                rc = cdc_hdr_flush(cdc, 1);
              }
              if (0 == rc)
              {
                cdc->map_distance 
                  = cdc->map.tbl.blocks      /* blocks used by map table */
                    + cdc->bat.tbl.blocks    /* blocks used by BAT       */
                    + cdc->bat.tbl.entries;  /* blocks used by data      */
              } else
              {
                cdc->env->close(cdc->fd);
              }
            }
            if (0 != rc)
            {
              free(cdc->filename);
            }
          }
        }
        if (0 != rc)
        {
          fd_table_entry_free(fd);
        }
      }
    }
    if (0 != rc)
    {
      fd = -1;
      errno = rc;
      if (0 != env_local)
      {
        comprio_env_free(env);
      }
    }
  }
  return fd;
}

/*
 * NAME
 *   comprio_seek
 *
 * DESCRIPTION
 *   seek into the file
 *
 * PARAMETERS
 *   fd     : file descriptor
 *   offset : offset
 *   whence : how (beginning, current, end)
 *
 * RETURN
 *   new position
 *   -1 for error, errno set appropriatly
 *     EBADF  : bad file descriptor
 *     EINVAL : whence is invalid, or file offset would go negative
 *
 * NOTES
 */
off_t comprio_seek(int fd, off_t offset, int whence)
{
  int    rc;
  cdc_t *cdc;
  off_t  pos;

  rc = fd_table_entry_to_cdc(fd, &cdc);
  if (0 != rc)
  {
    pos = -1;
  } else
  {
    switch (whence)
    {
      case SEEK_SET : /* from beginning        */
        if (offset < 0)
        {
          pos   = -1;
          errno = EINVAL;
        } else
        {
          cdc->offset = offset;
          pos         = cdc->offset;
        }
        break;
      case SEEK_CUR : /* from current position */
        if ((offset < 0) && (offset > cdc->offset))
        {
          pos = -1;
          errno = EINVAL;
        } else
        {
          cdc->offset += offset;
          pos          = cdc->offset;
        }
        break;
      case SEEK_END : /* from end of file      */
        if (offset > cdc->hdr.file_size)
        {
          pos = -1;
          errno = EINVAL;
        } else
        {
          cdc->offset = cdc->hdr.file_size - offset;
          pos         = cdc->offset;
        }
        break;

      default:
        pos   = -1;
        errno = EINVAL;
        break;
    }
  }
  return pos;
}

/*
 * NAME
 *   comprio_flush
 *
 * DESCRIPTION
 *   flush the buffers
 *
 * PARAMETERS
 *   fd : file descriptor
 *
 * RETURN
 *
 * NOTES
 */
int comprio_sync(int fd)
{
  int    rc;
  cdc_t *cdc;
  
  rc = fd_table_entry_to_cdc(fd, &cdc);
  if (0 == rc)
  {
    int rc1;

    if (cdc->hdr.dirty)
    {
      rc1 = cdc_data_flush(cdc);
      if (0 == rc)
      {
        rc = rc1;
      }
      rc1 = cdc_map_flush(cdc);
      if (0 == rc)
      {
        rc = rc1;
      }
      rc1 = cdc_bat_flush(cdc);
      if (0 == rc)
      {
        rc = rc1;
      }
      cdc->hdr.dirty = 0;
      rc1 = cdc_hdr_flush(cdc, 1);
      if (0 == rc)
      {
        rc = rc1;
      }
    }
  }
  return rc;
}

/*
 * NAME
 *   comprio_close
 *
 * DESCRIPTION
 *   close an open file
 *
 * PARAMETERS
 *   fd : file descriptor
 *
 * RETURN
 *
 * NOTES
 */
int comprio_close(int fd)
{
  int    rc;
  cdc_t *cdc;

  rc = fd_table_entry_to_cdc(fd, &cdc);
  if (0 == rc)
  {
    rc = comprio_sync(fd);
    cdc->env->close(cdc->fd);
    cdc_cleanup(cdc);
    comprio_env_free(cdc->env);
    fd_table_entry_free(fd);
  }
  return rc;
}

/*
 * NAME
 *   comprio_write
 *
 * DESCRIPTION
 *   write to a file
 *
 * PARAMETERS
 *   fd    : file descriptor
 *   buf   : buffer
 *   count : number of bytes to write
 *
 * RETURN
 *   number of bytes written, -1 on i/o error
 *
 * NOTES
 */
ssize_t comprio_write(int fd, const void *buf, size_t icount)
{
  int    rc;
  cdc_t *cdc;

  rc = fd_table_entry_to_cdc(fd, &cdc);
  if (0 != rc)
  {
    icount = -1;
  } else
  {
    const char   *buf_ptr;
    unsigned long cluster; /* cluster number        */
    unsigned long offset;  /* offset within cluster */
    size_t        count;

    buf_ptr = buf;
    count   = icount;

    cluster = cdc->offset / cdc->data.size;
    offset  = cdc->offset % cdc->data.size;

    while (count)
    {
      size_t        xfer;   /* number of bytes to transfer */

      if (offset || (count < cdc->data.size))
      {
        /*
         * partial block (ick!)
         */
        cdc_data_read(cdc, cluster);
        xfer = cdc->data.size - offset;
        if (count < xfer)
        {
          xfer = count;
        }
      } else
      {
        cdc_data_flush(cdc);
        xfer = (count < cdc->data.size)
          ? count
          : cdc->data.size;
        cdc->data.cluster.cluster = cluster;
      }

      memcpy(
        cdc->data.cluster.buffer + offset,
        buf_ptr,
        xfer
        );
      cdc->data.cluster.dirty = 1;
      cdc_hdr_dirty(cdc);
      buf_ptr += xfer;
      count   -= xfer;
      cluster++;
      offset = 0;
    }
    cdc->offset += icount;
    if (cdc->offset > cdc->hdr.file_size)
    {
      cdc->hdr.file_size = cdc->offset;
    }
  }
  return icount;
}

/*
 * NAME
 *   comprio_read
 *
 * DESCRIPTION
 *   read from a file
 *
 * PARAMETERS
 *   fd    : file descriptor
 *   buf   : buffer
 *   count : number of bytes to read
 *
 * RETURN
 *   number of bytes read
 *
 * NOTES
 */
ssize_t comprio_read(int fd, void *buf, size_t icount)
{
  int    rc;
  cdc_t *cdc;
  
  rc = fd_table_entry_to_cdc(fd, &cdc);
  if (0 != rc)
  {
    icount = -1;
  } else
  {
    char         *buf_ptr;
    unsigned long cluster;  /* cluster number        */
    unsigned long offset;   /* offset within cluster */
    unsigned long count;

    buf_ptr = buf;

    cluster = cdc->offset / cdc->data.size;
    offset  = cdc->offset % cdc->data.size;
    if (cdc->offset + icount > cdc->hdr.file_size)
    {
      if (cdc->offset > cdc->hdr.file_size)
      {
        icount = 0;
      } else
      {
        icount = cdc->hdr.file_size - cdc->offset;
      }
    }
    count = icount;
    cdc->offset += count;

    while (count)
    {
      size_t xfer;   /* number of bytes to transfer */

      cdc_data_read(cdc, cluster);
      xfer = cdc->data.size - offset;
      if (count < xfer)
      {
        xfer = count;
      }
      memcpy(
        buf_ptr, 
        cdc->data.cluster.buffer + offset,
        xfer
        );
      buf_ptr += xfer;
      count   -= xfer;
      cluster++;
      offset = 0;
    }
  } 
  return icount;
}


/*
 * NAME
 *   comprio_stat
 *
 * DESCRIPTION
 *   return info about the file
 *
 * PARAMETERS
 *   fd  : file descriptor
 *   buf : stat buf
 *
 * RETURN
 *
 * NOTES
 */
int comprio_stat(int fd, struct stat *buf)
{
  cdc_t *cdc;
  int    rc;

  rc = fd_table_entry_to_cdc(fd, &cdc);
  if (0 == rc)
  {
    rc = cdc->env->stat(cdc->fd, buf);
    if (0 == rc)
    {
      buf->st_size    = cdc->hdr.file_size;
      buf->st_blksize = cdc->blk.size;
      buf->st_blocks  = cdc->blk.blocks;
    }
  }
  return rc;
}

/*
 * NAME
 *   comprio_counter
 *
 * DESCRIPTION
 *   return data counters
 *
 * PARAMETERS
 *   fd  : file descriptor
 *   buf : [out] holds result
 *
 * RETURN
 *
 * NOTES
 */
int comprio_counter(int fd, comprio_stat_t *buf)
{
  cdc_t *cdc;
  int    rc;

  rc = fd_table_entry_to_cdc(fd, &cdc);
  if (0 == rc)
  {
    cdc_blk_counters_get(cdc, buf);
    cdc_data_counters_get(cdc, buf);
    cdc_bat_counters_get(cdc, buf);
    cdc_map_counters_get(cdc, buf);
  }
  return rc;
}

int comprio_counter_dump(int fd, FILE *out)
{
  cdc_t         *cdc;
  int            rc;

  rc = fd_table_entry_to_cdc(fd, &cdc);
  if (0 == rc)
  {
    comprio_stat_t st;
    struct stat    fst;

    rc = comprio_stat(fd, &fst);
    rc = comprio_counter(fd, &st);
    fprintf(
      out,
      "%s (%lu bytes)-->\n"
      "  block size(%lu) blocks(%lu) read(%lu) write(%lu)\n"
      "  data  size(%lu) blocks(%lu) read(%lu) write(%lu)\n"
      "  BAT   blocks(%lu) count(%lu) read(%lu) write(%lu)\n"
      "  map   blocks(%lu) count(%lu) read(%lu) write(%lu)\n",
      cdc->filename,
      fst.st_size,
      st.blk.size, st.blk.blocks, st.blk.counters.read, st.blk.counters.write,
      st.data.size, st.data.blocks, st.data.counters.read, st.data.counters.write,
      st.bat.blocks, st.bat.count, st.bat.counters.read, st.bat.counters.write,
      st.map.blocks, st.map.count, st.map.counters.read, st.map.counters.write
      );
  }
  return rc;
}

/*
 * NAME
 *   comprio_lock
 *
 * DESCRIPTION
 *   lock the file
 *
 * PARAMETERS
 *   fd   : file descriptor
 *   type : lock type
 *
 * RETURN
 *
 * NOTES
 */
int comprio_lock(int fd, int type)
{
  cdc_t *cdc;
  int    rc;

  rc = fd_table_entry_to_cdc(fd, &cdc);
  if (0 == rc)
  {
    rc = flock(cdc->fd, type);
  }
  return rc;
}

