#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <assert.h>

#include "cdc.h"

/*
 * NAME
 *   bat_new_table
 *
 * DESCRIPTION
 *   called when a newly initialzed bat table is written
 *
 * PARAMETERS
 *   cdc : compressed file descriptor
 *
 * RETURN
 *
 * NOTES
 *   when before a new bat table is written
 *   the preceding map table page must also be written
 */
int cdc_bat_tbl_create(cdc_t *cdc)
{
  int rc = 0;

  if (cdc->bat.tbl.count == cdc->map.tbl.count)
  {
    rc = cdc_map_tbl_create(cdc);
  }
  if (0 == rc)
  {
    rc = cdc_tbl_create(cdc, &cdc->bat.tbl);
  }

  return rc;
}

static int bat_tbl_create(cdc_t *cdc, cdc_tbl_t *tbl)
{
  return cdc_bat_tbl_create(cdc);
}

/*
 * NAME
 *   cdc_bat_init
 *
 * DESCRIPTION
 *   initialize the bat elements in the cdc structure
 *
 * PARAMETERS
 *   cdc : initialized compressed file descriptor
 *
 * RETURN
 *   0      : no error
 *   ENOMEM
 *
 * NOTES
 */
int cdc_bat_init(cdc_t *cdc, unsigned long blocks)
{
  /*
   * initialize the bat cluster
   */
  cdc->bat.count      = 0;
  cdc->bat.first_free = BAT_ENTRY_EOL;
  return cdc_tbl_init(
    cdc,
    &cdc->bat.tbl,
    blocks, 
    BAT_ENTRY_EOL,
    cdc->map.tbl.blocks, /* bat follows map */
    bat_tbl_create
    );
}

/*
 * NAME
 *   cdc_bat_cleanup
 *
 * DESCRIPTION
 *   cleanup the bat elements in the cdc
 *
 * PARAMETERS
 *   cdc : initialized compressed file descriptor
 *
 * RETURN
 *   0 :  no error
 *
 * NOTES
 */
int cdc_bat_cleanup(cdc_t *cdc)
{
  return cdc_tbl_cleanup(cdc, &cdc->bat.tbl);
}

int cdc_bat_hdr_write(cdc_t *cdc, char *ptr, size_t *sz)
{
  if (ptr)
  {
    U32_PUT(ptr, cdc->bat.tbl.blocks);
    U32_PUT(ptr, cdc->bat.tbl.count);
    U32_PUT(ptr, cdc->bat.count);
    U32_PUT(ptr, cdc->bat.first_free);
  }
  *sz = 16;
  return 0;
}

int cdc_bat_hdr_read(cdc_t *cdc, const char *ptr, const char **eptr)
{
  unsigned long blocks;
  int           rc;

  U32_GET(blocks, ptr);
  rc = cdc_bat_init(cdc, blocks);
  if (0 == rc)
  {
    U32_GET(cdc->bat.tbl.count,  ptr);
    U32_GET(cdc->bat.count,      ptr);
    U32_GET(cdc->bat.first_free, ptr);
    *eptr= ptr;
  }
  return rc;
}

/*
 * NAME
 *   cdc_bat_flush
 *
 * DESCRIPTION
 *   flush the current bat cluster if dirty
 *
 * PARAMETERS
 *   cdc   : initialized compressed file descriptor
 *
 * RETURN
 *   0   : no error
 *   EIO : I/O error
 *
 * NOTES
 */
int cdc_bat_flush(
  cdc_t *cdc
  )
{
  return cdc_tbl_flush(cdc, &cdc->bat.tbl);
}

/*
 * NAME
 *   cdc_bat_entry_get
 *
 * DESCRIPTION
 *   return the value at the bat entry
 *
 * PARAMETERS
 *   cdc   : initialized compressed file descriptor
 *   entry : entry number
 *   value : [out] holds result on success
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
int cdc_bat_entry_get(
  cdc_t         *cdc, 
  unsigned long  entry, 
  unsigned long *value
  )
{
  return cdc_tbl_entry_get(cdc, &cdc->bat.tbl, entry, value);
}

/*
 * NAME
 *   cdc_bat_entry_set
 *
 * DESCRIPTION
 *   set a bat table entry
 *
 * PARAMETERS
 *   cdc   : initialized compressed file descriptor
 *   entry : entry number
 *   value : value to set
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
int cdc_bat_entry_set(
  cdc_t         *cdc,
  unsigned long  entry,
  unsigned long  value
  )
{
  return cdc_tbl_entry_set(cdc, &cdc->bat.tbl, entry, value);
}

/*
 * NAME
 *   cdc_bat_entry_alloc
 *
 * DESCRIPTION
 *   return an unused bat entry
 *
 * PARAMETERS
 *   cdc   : initialzed compress filed descriptor
 *   value : [out] holds result on success
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 *   
 */
int cdc_bat_entry_alloc(
  cdc_t         *cdc,
  unsigned long *value
  )
{
  int           rc;
  unsigned long ety;

  ety = cdc->bat.first_free;
  if (BAT_ENTRY_EOL == ety)
  {
    ety = cdc->bat.count;
    cdc->bat.count++;
    rc = 0;
  } else
  {
    rc = cdc_bat_entry_get(cdc, ety, &cdc->bat.first_free);
  }
  if (0 == rc)
  {
    rc = cdc_bat_entry_set(cdc, ety, BAT_ENTRY_EOL);
    if (0 == rc)
    {
      *value = ety;
    }
  }
  return rc;
}

/*
 * NAME
 *   cdc_bat_entry_free
 *
 * DESCRIPTION
 *   add a bat entry to the free list
 *
 * PARAMETERS
 *   cdc   :
 *   entry : entry # to free
 *
 * RETURN
 *   possible I/O error from table routines
 *
 * NOTES
 *   this is not very common. it is called when an data cluster compresses
 *   to fewer blocks than previously. i'm going to keep this list ordered.
 */
int cdc_bat_entry_free(cdc_t *cdc, unsigned long entry)
{
  unsigned long eptr;
  unsigned long eptr_pv;
  int           rc;

  if (BAT_ENTRY_EOL == cdc->bat.first_free)
  {
    eptr_pv = BAT_ENTRY_EOL;
    eptr    = BAT_ENTRY_EOL;
    rc      = 0;
  } else
  {
    for (
      eptr_pv = BAT_ENTRY_EOL,
        eptr = cdc->bat.first_free,
        rc = 0;
      (0 == rc)
      && (entry > eptr);
      eptr_pv = eptr,
        rc = cdc_bat_entry_get(cdc, eptr, &eptr)
        )
      ;
  }
  if (0 == rc)
  {
    assert(eptr != entry);
    rc = cdc_bat_entry_set(cdc, entry, eptr);
    if (0 == rc)
    {
      if (BAT_ENTRY_EOL == eptr_pv)
      {
        cdc->bat.first_free = entry;
      } else
      {
        rc = cdc_bat_entry_set(cdc, eptr_pv, entry);
      }
    }
  }
  return rc;
}

/*
 * NAME
 *   bat_counters_get
 *
 * DESCRIPTION
 *   return the bat counters
 *
 * PARAMETERS
 *   cdc : 
 *   buf : [out] holds result
 *
 * RETURN
 *   0
 *
 * NOTES
 */
int cdc_bat_counters_get(cdc_t *cdc, comprio_stat_t *buf)
{
  buf->bat.counters = cdc->bat.tbl.counter;
  buf->bat.blocks   = cdc->bat.tbl.blocks;
  buf->bat.count    = cdc->bat.tbl.count;
  return 0;
}

