#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include "cdc.h"

/*
 * NAME
 *   cdc_blk_init
 *
 * DESCRIPTION
 *   initialize the blk structure in the cdc header
 *
 * PARAMETERS
 *   cdc  : cdc header
 *   size : block size
 *
 * RETURN
 *
 * NOTES
 */
int cdc_blk_init(cdc_t *cdc, unsigned long size)
{
  memset(&cdc->blk, 0, sizeof(cdc->blk));
  cdc->blk.size   = size;
  return 0;
}

/*
 * NAME
 *   cdc_blk_cleanup
 *
 * DESCRIPTION
 *   cleanup the blk structure in the cdc header
 *
 * PARAMETERS
 *   cdc : cdc 
 *
 * RETURN
 *
 * NOTES
 *   here for symmetry
 */
int cdc_blk_cleanup(cdc_t *cdc)
{
  return 0;
}

/*
 * NAME
 *   cdc_blk_hdr_read
 *
 * DESCRIPTION
 *   read the block elements of the header
 *
 * PARAMETERS
 *   cdc   : cdc entry
 *   ptr   : start of data to read
 *   entry : [out] points to first unused character
 *
 * RETURN
 *
 * NOTES
 */
int cdc_blk_hdr_read(cdc_t *cdc, const char *ptr, const char **eptr)
{
  unsigned long size;
  int           rc;

  U32_GET(size, ptr);
  rc = cdc_blk_init(cdc, size);
  if (0 == rc)
  {
    U32_GET(cdc->blk.blocks, ptr);
    *eptr = ptr;
  }
  return rc;
}
  
/*
 * NAME
 *   cdc_blk_hdr_write
 *
 * DESCRIPTION
 *   write the block elements to the header
 *
 * PARAMETERS
 *   cdc  : cdc
 *   ptr  : place to write
 *   eptr : first unused position
 *
 * RETURN
 *
 * NOTES
 */
int cdc_blk_hdr_write(cdc_t *cdc, char *ptr, size_t *sz)
{
  if (ptr)
  {
    U32_PUT(ptr, cdc->blk.size);
    U32_PUT(ptr, cdc->blk.blocks);
  }
  *sz = 8;
  return 0;
}
  
/*
 * NAME
 *   block_seek
 *
 * DESCRIPTION
 *   set the file pointer to the beginning of a block
 *
 * PARAMETERS
 *   cdc   : compressed data descriptor
 *   block : block number
 *
 * RETURN
 *   0   : no error
 *   EIO : seek error
 *
 * NOTES
 */
static int block_seek(cdc_t *cdc, unsigned long block)
{
  return (-1 == cdc->env->seek(cdc->fd, block * cdc->blk.size, SEEK_SET)) 
    ? EIO
    : 0;
}

/*
 * NAME
 *   cdc_blk_read
 *
 * DESCRIPTION
 *   read a file block
 *
 * PARAMETERS
 *   cdc    : compressed file descriptor
 *   block  : block number
 *   buf    : data
 *   count  : number of blocks
 *   result : [out] number of blocks read
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
int cdc_blk_read(
  cdc_t         *cdc, 
  unsigned long  block, 
  void          *buf,
  unsigned long  count,
  unsigned long *result
  )
{
  int rc;

  rc = block_seek(cdc, block);
  if (0 == rc)
  {
    ssize_t ct;

    ct = cdc->env->read(
      cdc->fd, 
      buf, 
      cdc->blk.size * count
      );
    if (-1 == ct)
    {
      rc = errno;
    } else
    {
      ct /= cdc->blk.size;
      *result = ct;
      cdc->blk.counter.read += ct;
    }
  }
  return rc;
}

/*
 * NAME
 *   cdc_blk_write
 *
 * DESCRIPTION
 *   write blocks to a file
 *
 * PARAMETERS
 *   cdc : compressed file descriptor
 *   block : block number
 *   buf   : data to write
 *   count : number of blocks
 *
 * RETURN
 *
 * NOTES
 */
int cdc_blk_write(
  cdc_t         *cdc,
  unsigned long  block, 
  const void    *buf,
  unsigned long  count
  )
{
  int rc;

  rc = block_seek(cdc, block);
  if (0 == rc)
  {
    ssize_t ct;

    ct = cdc->env->write(cdc->fd, buf, cdc->blk.size * count);
    if (-1 == ct)
    {
      rc = errno;
    } else
    {
      ct /= cdc->blk.size;
      assert(ct == count);
      cdc->blk.counter.write += ct;
      if (block + ct > cdc->blk.blocks)
      {
        cdc->blk.blocks = block + ct;
      }
    }
  }
  return rc;
}

int cdc_blk_counters_get(cdc_t *cdc, comprio_stat_t *buf)
{
  buf->blk.counters = cdc->blk.counter;
  buf->blk.size     = cdc->blk.size;
  buf->blk.blocks   = cdc->blk.blocks;
  return 0;
}

