#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <assert.h>

#include "cdc.h"

/*
 * NAME
 *   cdc_data_init
 *
 * DESCRIPTION
 *   initialize the data elements in the cdc
 *
 * PARAMETERS
 *   cdc : compressed descriptor block
 *
 * RETURN
 *   0      : no error
 *   ENOMEM : out of memory
 *
 * NOTES
 */
int cdc_data_init(cdc_t *cdc, unsigned long blocks)
{
  int rc;

  cdc->data.blocks          = blocks;
  cdc->data.size            = cdc->blk.size * blocks;
  cdc->data.cluster.dirty   =  0;
  cdc->data.cluster.cluster = -1;
  cdc->data.cluster.buffer  = malloc(cdc->data.size);
  cdc->data.zbuffer         = malloc(cdc->data.size);

  if ((0 == cdc->data.cluster.buffer)
    || (0 == cdc->data.zbuffer))
  {
    rc = ENOMEM;
  } else
  {
    rc = cdc->env->inflate.init(cdc->env->inflate.arg);
    if (0 == rc)
    {
      rc = cdc->env->deflate.init(cdc->env->deflate.arg);
      if (0 != rc)
      {
        cdc->env->inflate.cleanup(cdc->env->inflate.arg);
      }
    }
  }
  if (0 != rc)
  {
    free(cdc->data.zbuffer);
    free(cdc->data.cluster.buffer);
  }
  return rc;
}

/*
 * NAME
 *   cdc_data_cleanup
 *
 * DESCRIPTION
 *   cleanup the data elements in the cdc
 *
 * PARAMETERS
 *   cdc : compressed descriptor block
 *
 * RETURN
 *   0 : no error
 *
 * NOTES
 */
int cdc_data_cleanup(cdc_t *cdc)
{
  cdc->env->deflate.cleanup(cdc->env->deflate.arg);
  cdc->env->inflate.cleanup(cdc->env->inflate.arg);
  free(cdc->data.zbuffer);
  free(cdc->data.cluster.buffer);
  return 0;
}

int cdc_data_hdr_read(cdc_t *cdc, const char *ptr, const char **eptr)
{
  unsigned long blocks;

  U32_GET(blocks, ptr);
  *eptr = ptr;
  return cdc_data_init(cdc, blocks);
}

int cdc_data_hdr_write(cdc_t *cdc, char *ptr, size_t *sz)
{
  if (ptr)
  {
    U32_PUT(ptr, cdc->data.blocks);
  }
  *sz = 4;
  return 0;
}

/*
 * NAME
 *   data_block_to_file_block
 *
 * DESCRIPTION
 *   given a data block number, return the file block number
 *
 * PARAMETERS
 *   cdc     : cdc
 *   block   : data block number
 *   dst     : [out] holds result
 *
 * RETURN
 *   0   : no error
 *
 * NOTES
 */
static int data_block_to_file_block(
  cdc_t         *cdc,
  unsigned long  block,
  unsigned long *dst
  )
{
  *dst 
    = 1  /* header */
      + (block / cdc->bat.tbl.entries) * (cdc->map_distance)
      + cdc->bat.tbl.blocks
      + cdc->map.tbl.blocks
      + (block % cdc->bat.tbl.entries);
  return 0;
}

/*
 * NAME
 *   data_block_write
 *
 * DESCRIPTION
 *   write a data block
 *
 * PARAMETERS
 *   cdc     : compressed file descriptor
 *   blockno : block number
 *   buffer  : block buffer
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
static int data_block_write(
  cdc_t        *cdc,
  unsigned long blockno,
  const void   *buffer
  )
{
  data_block_to_file_block(cdc, blockno, &blockno);
  cdc->data.counter.write++;
  return cdc_blk_write(
    cdc,
    blockno,
    buffer,
    1
    );
}

/*
 * NAME
 *   data_block_read
 *
 * DESCRIPTION
 *   read a data block
 *
 * PARAMETERS
 *   cdc    : compressed file descriptor
 *   block  : block to read
 *   buffer : buffer to holds data
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
static int data_block_read(
  cdc_t         *cdc, 
  unsigned long  block, 
  void          *buffer
  )
{
  unsigned long count;

  data_block_to_file_block(cdc, block, &block);
  cdc->data.counter.read++;
  return cdc_blk_read(
    cdc,
    block,
    buffer,
    1,
    &count
    );
}

/*
 * NAME
 *   cdc_data_flush
 *
 * DESCRIPTION
 *   flush a data cluster if dirty
 *
 * PARAMETERS
 *   cdc : compressed file descriptor
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
int cdc_data_flush(cdc_t *cdc)
{
  int rc;

  rc = 0;
  if (cdc->data.cluster.dirty)
  {
    unsigned long bat_entry;
    unsigned long bat_entry_next;
    const char   *ptr;
    unsigned long count;
    size_t        zsize;
    unsigned long compress_flag;

    cdc->data.cluster.dirty = 0;
    count      = cdc->data.blocks;
    ptr        = cdc->data.cluster.buffer;
    compress_flag = 0;

    zsize = cdc->data.size;
    rc = cdc->env->deflate.fn(
      cdc->env->deflate.arg,
      cdc->data.zbuffer,
      &zsize,
      cdc->data.cluster.buffer,
      cdc->data.size
      );
    if (0 == rc)
    {
      /*
       * success! but if we don't save at least 1 block, don't bother
       */
      count = (zsize + cdc->blk.size - 1) / cdc->blk.size;
      if (count < cdc->data.blocks)
      {
        compress_flag = MAP_FLAG_COMPRESSED;
        ptr = cdc->data.zbuffer;
      } 
    }
    rc = cdc_map_entry_get(cdc, cdc->data.cluster.cluster, &bat_entry);

    if (0 == rc)
    {
      if (BAT_ENTRY_EOL == bat_entry)
      {
        rc = cdc_bat_entry_alloc(cdc, &bat_entry); /* get a free bat entry */
        if (0 == rc)
        {
          rc = cdc_map_entry_set(
            cdc, 
            cdc->data.cluster.cluster, 
            bat_entry | compress_flag
            );
        }
      } else
      {
        unsigned long map_entry;

        map_entry = bat_entry;
        if (compress_flag)
        {
          map_entry |= MAP_FLAG_COMPRESSED;
        } else
        {
          map_entry &= ~MAP_FLAG_COMPRESSED;
        }
        if (bat_entry != map_entry)
        {
          rc = cdc_map_entry_set(cdc, cdc->data.cluster.cluster, map_entry);
        }
        bat_entry &= ~MAP_FLAG_COMPRESSED;
      }
    }
    while ((0 == rc) && count)
    {
      rc = data_block_write(cdc, bat_entry, ptr);
      if (0 == rc)
      {
        rc = cdc_bat_entry_get(cdc, bat_entry, &bat_entry_next);
      }
      if (0 == rc)
      {
        count--;
        if (count)
        {
          ptr += cdc->blk.size;

          if (0 == rc)
          {
            if (BAT_ENTRY_EOL == bat_entry_next)
            {
              rc = cdc_bat_entry_alloc(cdc, &bat_entry_next);
              if (0 == rc)
              {
                rc = cdc_bat_entry_set(cdc, bat_entry, bat_entry_next);
              }
            }
            bat_entry = bat_entry_next;
          }
        }
      }
    }
#if 1
    /*
     * free up any remaining entries
     *   --- occurs if the current cluster compresses to less than
     *       the previous one
     */
    if ((0 == rc) && (BAT_ENTRY_EOL != bat_entry_next))
    {
      cdc_bat_entry_set(cdc, bat_entry, BAT_ENTRY_EOL); /* mark the end */
      do
      {
        bat_entry = bat_entry_next;
        rc = cdc_bat_entry_get(cdc, bat_entry, &bat_entry_next);
        if (0 == rc)
        {
          rc = cdc_bat_entry_free(cdc, bat_entry);
        }
      } while ((0 == rc) && (BAT_ENTRY_EOL != bat_entry_next));
    }
#endif
  }
  return rc;
}

/*
 * NAME
 *   cdc_data_read
 *
 * DESCRIPTION
 *   read a data cluster
 *
 * PARAMETERS
 *   cdc     : descriptor
 *   cluster : cluster number
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
int cdc_data_read(cdc_t *cdc, unsigned long cluster)
{
  int           rc;
  
  rc = 0;
  if (cluster != cdc->data.cluster.cluster)
  {
    unsigned long bat_entry;
    char         *ptr;
    char *base;
    int   compressed;

    rc = cdc_data_flush(cdc);
    if (0 == rc)
    {
      cdc->data.cluster.cluster = cluster;
      
      ptr = cdc->data.cluster.buffer;

      rc = cdc_map_entry_get(cdc, cluster, &bat_entry);
      if (0 == rc)
      {
        if ((BAT_ENTRY_EOL != bat_entry) 
          && (bat_entry & MAP_FLAG_COMPRESSED))
        {
          compressed = 1;
          base       = cdc->data.zbuffer;
          bat_entry &= ~MAP_FLAG_COMPRESSED;
        } else
        {
          compressed = 0;
          base       = cdc->data.cluster.buffer;
        }
        ptr = base;
        while ((0 == rc) && (BAT_ENTRY_EOL != bat_entry))
        {
          assert(ptr - base < cdc->data.size);
          rc = data_block_read(cdc, bat_entry, ptr);
          if (0 == rc)
          {
            ptr += cdc->blk.size;
            rc = cdc_bat_entry_get(cdc, bat_entry, &bat_entry);
          }
        }
        if (0 == rc)
        {
          if (compressed)
          {
            size_t zsize;

            zsize = cdc->data.size;

            rc = cdc->env->inflate.fn(
              cdc->env->inflate.arg,
              cdc->data.cluster.buffer,
              &zsize,
              cdc->data.zbuffer,
              ptr - base
              );
            if (0 == rc)
            {
              assert(zsize == cdc->data.size);
            }
          } else
          {
            ptrdiff_t used;

            used = ptr - (char *) cdc->data.cluster.buffer;
            if (used < cdc->data.size)
            {
              memset(ptr, 0, cdc->data.size - used);
            }
          }
        }
      }
    }
  }
  return rc;
}
  
int cdc_data_counters_get(cdc_t *cdc, comprio_stat_t *buf)
{
  buf->data.counters = cdc->data.counter;
  buf->data.size     = cdc->data.size;
  buf->data.blocks   = cdc->data.blocks;
  return 0;
}


