#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include "cdc.h"

/*
 * NAME
 *   cdc_hdr_write
 *
 * DESCRIPTION
 *   write the file header
 *
 * PARAMETERS
 *   cdc      : compressed file descriptor
 *   complete : 0     : only flush file header
 *              non-0 : flush all subsystem headers
 *
 * RETURN
 *
 * NOTES
 */
int cdc_hdr_flush(cdc_t *cdc, int complete)
{
  char  *ptr;
  size_t sz;
  int    rc = 0;

  ptr = cdc->hdr.buffer;
  if (0 == ptr)
  {
    size_t sum = 20; /* size of bits i write */

    cdc_blk_hdr_write(cdc, 0, &sz);
    sum += sz;
    cdc_map_hdr_write(cdc, 0, &sz);
    sum += sz;
    cdc_bat_hdr_write(cdc, 0, &sz);
    sum += sz;
    cdc_data_hdr_write(cdc, 0, &sz);
    sum += sz;

    cdc->hdr.blocks = (sum + cdc->blk.size - 1) / cdc->blk.size;
    cdc->hdr.buffer = malloc(cdc->hdr.blocks * cdc->blk.size);
    ptr = cdc->hdr.buffer;
    if (0 == ptr)
    {
      rc = ENOMEM;
    }
  }
  if (0 == rc)
  {
    U32_PUT(ptr, cdc->hdr.magic);
    U32_PUT(ptr, cdc->hdr.version);
    U32_PUT(ptr, cdc->hdr.file_size);
    U32_PUT(ptr, cdc->hdr.blocks);
    U32_PUT(ptr, cdc->hdr.dirty);
    if (complete)
    {
      cdc_blk_hdr_write(cdc, ptr, &sz);
      ptr += sz;
      cdc_map_hdr_write(cdc, ptr, &sz);
      ptr += sz;
      cdc_bat_hdr_write(cdc, ptr, &sz);
      ptr += sz;
      cdc_data_hdr_write(cdc, ptr, &sz);
      ptr += sz;
    }
    rc = cdc_blk_write(
      cdc,
      0,
      cdc->hdr.buffer,
      cdc->hdr.blocks
      );
  }
  return rc;
}

/*
 * NAME
 *   cdc_hdr_read
 *
 * DESCRIPTION
 *   read the file header
 *
 * PARAMETERS
 *   cdc : compressed file descriptor
 *
 * RETURN
 *   0      : no error
 *   EINVAL : not a valid comprio file
 *
 * NOTES
 *   note the liberal use of goto's. code purests will
 *   probably balk, but i think it's MUCH easier to read
 *   than nesting if's (as the original code did). also,
 *   my editor shows gotos in red which makes them very
 *   obvious
 */
int cdc_hdr_read(cdc_t *cdc)
{
  char           hdr[CDC_HEADER_SIZE];
  const char    *ptr;
  int            rc;
  unsigned long  ct;

  rc = cdc_blk_init(cdc, CDC_HEADER_SIZE);
  if (0 != rc)
  {
    goto error;
  }
  rc = cdc_blk_read(cdc, 0, hdr, 1, &ct);
  if (0 != rc)
  {
    goto unwind_blk;
  }
  if (1 != ct)
  {
    rc = EINVAL;
    goto unwind_blk;
  } 
  ptr = hdr;
  U32_GET(cdc->hdr.magic,     ptr);
  U32_GET(cdc->hdr.version,   ptr);
  U32_GET(cdc->hdr.file_size, ptr);
  U32_GET(cdc->hdr.blocks,    ptr);
  U32_GET(cdc->hdr.dirty,     ptr);
  if ((cdc->hdr.magic != COMPRIO_MAGIC)
    || (cdc->hdr.version != COMPRIO_VERSION)
    || (0 != cdc->hdr.dirty))
  {
    rc = EINVAL;
    goto unwind_blk;
  }
  cdc_blk_cleanup(cdc);
  rc = cdc_blk_hdr_read(cdc, ptr, &ptr);
  if (0 != rc)
  {
    goto error;
  }
  cdc->hdr.buffer = malloc(cdc->hdr.blocks * cdc->blk.size);
  if (0 == cdc->hdr.buffer)
  {
    rc = ENOMEM;
    goto unwind_blk;
  }
  rc = cdc_blk_read(cdc, 0, cdc->hdr.buffer, cdc->hdr.blocks, &ct);
  if (0 != rc)
  {
    goto unwind_blk;
  }
  if (ct != cdc->hdr.blocks)
  {
    rc = EIO;
    goto unwind_blk;
  } 
  ptr = cdc->hdr.buffer + (ptr - hdr);
  rc = cdc_map_hdr_read(cdc, ptr, &ptr);
  if (0 != rc)
  {
    goto unwind_blk;
  }
  rc = cdc_bat_hdr_read(cdc, ptr, &ptr);
  if (0 != rc)
  {
    goto unwind_map;
  }
  rc = cdc_data_hdr_read(cdc, ptr, &ptr);
  if (0 != rc)
  {
    goto unwind_bat;
  }
  cdc->map_distance 
    = cdc->map.tbl.blocks
      + cdc->bat.tbl.blocks
      + cdc->bat.tbl.entries;
  return rc;
unwind_bat:
  cdc_bat_cleanup(cdc);
unwind_map:
  cdc_map_cleanup(cdc);
unwind_blk:
  cdc_blk_cleanup(cdc);
  free(cdc->hdr.buffer);
error:
  return rc;
}
  
/*
 * NAME
 *   cdc_hdr_dirty
 *
 * DESCRIPTION
 *   mark the file as dirty
 *
 * PARAMETERS
 *   cdc : descriptoy
 *
 * RETURN
 *   none
 *
 * NOTES
 *   if this is the first time the file has been marked dirty,
 *   flush the header with the dirty bit set
 *   the file remains dirty until comprio_sync() is called and
 *   successful!
 */
int cdc_hdr_dirty(cdc_t *cdc)
{
  int rc;

  if (0 == cdc->hdr.dirty)
  {
    cdc->hdr.dirty = 1;
    /*
     * flush the header
     */
    rc = cdc_hdr_flush(cdc, 0);
  } else
  {
    rc = 0;
  }
  return rc;
}

