#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include "cdc.h"

/*
 * NAME
 *   map_table_create
 *
 * DESCRIPTION
 *   create any missing map and bat
 *
 * PARAMETERS
 *   cdc    : compressed file descriptor
 *   tbl    : table
 *   tbl_no : table number to create
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
int cdc_map_tbl_create(cdc_t *cdc)
{
  int rc;

  if (cdc->bat.tbl.count < cdc->map.tbl.count)
  {
    rc = cdc_bat_tbl_create(cdc);
  }
  return cdc_tbl_create(cdc, &cdc->map.tbl);
}

static int map_tbl_create(cdc_t *cdc, cdc_tbl_t *tbl)
{
  return cdc_map_tbl_create(cdc);
}

/*
 * NAME
 *   cdc_map_init
 *
 * DESCRIPTION
 *   initialize the map elements of a cdc
 *
 * PARAMETERS
 *   cdc : initialized compressed file descriptor
 *
 * RETURN
 *   0      : no error
 *   ENOMEM : out of memory
 *
 * NOTES
 */
int cdc_map_init(cdc_t *cdc, unsigned long blocks)
{
  return cdc_tbl_init(
    cdc,
    &cdc->map.tbl,
    blocks,
    BAT_ENTRY_EOL,
    0,
    map_tbl_create
    );
}

/*
 * NAME
 *   cdc_map_cleanup
 *
 * DESCRIPTION
 *   cleanup the map elements of a cdc
 *
 * PARAMETERS
 *   cdc : initialized compressed file descriptor
 *
 * RETURN
 *   0 : no error
 *
 * NOTES
 */
int cdc_map_cleanup(cdc_t *cdc)
{
  return cdc_tbl_cleanup(cdc, &cdc->map.tbl);
}

int cdc_map_hdr_read(cdc_t *cdc, const char *ptr, const char **eptr)
{
  unsigned long blocks;
  int           rc;

  U32_GET(blocks, ptr);
  rc = cdc_map_init(cdc, blocks);
  if (0 == rc)
  {
    U32_GET(cdc->map.tbl.count, ptr);
    *eptr = ptr;
  }
  return rc;
}

int cdc_map_hdr_write(cdc_t *cdc, char *ptr, size_t *sz)
{
  if (ptr)
  {
    U32_PUT(ptr, cdc->map.tbl.blocks);
    U32_PUT(ptr, cdc->map.tbl.count);
  }
  *sz = 8;
  return 0;
}

/*
 * NAME
 *   cdc_map_flush
 *
 * DESCRIPTION
 *   flush the current map cluster if dirty
 *
 * PARAMETERS
 *   cdc : initialized compressed file descriptor
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
int cdc_map_flush(
  cdc_t *cdc
  )
{
  return cdc_tbl_flush(cdc, &cdc->map.tbl);
}

/*
 * NAME
 *   cdc_map_entry_get
 *
 * DESCRIPTION
 *   return a map table entry
 *
 * PARAMETERS
 *   cdc   : initialized compressed file descriptor
 *   entry : entry number
 *   value : [out] holds result on success
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
int cdc_map_entry_get(
  cdc_t         *cdc,
  unsigned long  entry,
  unsigned long *value
  )
{
  return cdc_tbl_entry_get(cdc, &cdc->map.tbl, entry, value);
}

/*
 * NAME
 *   cdc_map_entry_set
 *
 * DESCRIPTION
 *   set a map table entry
 *
 * PARAMETERS
 *   cdc   : compressed file descriptor
 *   entry : entry number
 *   value : value to set to
 *
 * RETURN
 *   0   : no error
 *   EIO : i/o error
 *
 * NOTES
 */
int cdc_map_entry_set(
  cdc_t        *cdc,
  unsigned long entry,
  unsigned long value
  )
{
  return cdc_tbl_entry_set(cdc, &cdc->map.tbl, entry, value);
}

int cdc_map_counters_get(cdc_t *cdc, comprio_stat_t *buf)
{
  buf->map.counters = cdc->map.tbl.counter;
  buf->map.blocks   = cdc->map.tbl.blocks;
  buf->map.count    = cdc->map.tbl.count;
  return 0;
}


