#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "cdc.h"

int cdc_tbl_init(
  cdc_t              *cdc,
  cdc_tbl_t          *tbl, 
  unsigned long       blocks,
  unsigned long       phantom_value,
  unsigned long       block_offset,
  cdc_tbl_create_fn_t create_fn
  )
{
  tbl->count = 0;
  tbl->entries       
    = blocks 
      * cdc->blk.size 
      / sizeof(tbl->buffer.data[0]);
  tbl->phantom_value  = phantom_value;
  tbl->block_offset   =  block_offset;
  tbl->blocks         =  blocks;
  tbl->is_phantom     = 0;
  tbl->create_fn      = create_fn;
  tbl->buffer.dirty   =  0;
  tbl->buffer.block   = -1;
  tbl->buffer.data    = calloc(blocks, cdc->blk.size);

  return (0 == tbl->buffer.data)
    ? ENOMEM
    : 0;
}

int cdc_tbl_cleanup(cdc_t *cdc, cdc_tbl_t *tbl)
{
  free(tbl->buffer.data);
  return 0;
}

int cdc_tbl_flush(cdc_t *cdc, cdc_tbl_t *tbl)
{
  int rc;

  rc = 0;
  if (0 != tbl->buffer.dirty)
  {
    tbl->buffer.dirty = 0;
    rc = cdc_blk_write(
      cdc,
      tbl->buffer.block,
      tbl->buffer.data,
      tbl->blocks
      );
    if (0 == rc)
    {
      tbl->counter.write++;
    }
  }
  return rc;
}

static int tbl_sync(
  cdc_t         *cdc,
  cdc_tbl_t     *tbl,
  unsigned long  entry
  )
{
  int           rc;
  unsigned long tbl_no;
  unsigned long tbl_block;

  rc = 0;
  tbl_no = entry / tbl->entries; /* table # */
  tbl_block
    = 1 
    + tbl->block_offset
    + tbl_no * cdc->map_distance;
  if (tbl_block != tbl->buffer.block)
  {
    rc = cdc_tbl_flush(cdc, tbl);
    if (0 == rc)
    {
      unsigned long blocks;

      tbl->buffer.block = tbl_block;
      rc = cdc_blk_read(
        cdc,
        tbl->buffer.block,
        tbl->buffer.data,
        tbl->blocks,
        &blocks
        );
      if (0 == rc)
      {
        if (0 == blocks)
        {
          tbl->is_phantom = 1;
        } else if (blocks != tbl->blocks)
        {
          rc = EIO;
        } else
        {
          tbl->is_phantom = 0;
          tbl->counter.read++;
        }
      }
    }
  }
  return rc;
}

int cdc_tbl_entry_get(
  cdc_t         *cdc, 
  cdc_tbl_t     *tbl,
  unsigned long  entry, 
  unsigned long *value
  )
{
  int rc;
  
  rc = tbl_sync(cdc, tbl, entry);
  if (0 == rc)
  {
    *value = (tbl->is_phantom)
      ? tbl->phantom_value
      : tbl->buffer.data[entry % tbl->entries];
  }
  return rc;
}

int cdc_tbl_entry_set(
  cdc_t         *cdc,
  cdc_tbl_t     *tbl,
  unsigned long  entry,
  unsigned long  value
  )
{
  int           rc;
  unsigned long value_pv;
  
  rc = cdc_tbl_entry_get(cdc, tbl, entry, &value_pv);

  /*
   * note : if setting the phantom value to a phantom table
   *        it must be physically set in order to correctly
   *        initialize the table
   */
  if ((0 == rc) 
    && ((value_pv != value) || tbl->is_phantom))
  {
    if (tbl->is_phantom)
    {
      tbl->is_phantom = 0;
      rc = tbl->create_fn(cdc, tbl);
    }
    if (0 == rc)
    {
      tbl->buffer.data[entry % tbl->entries] = value;
      tbl->buffer.dirty = 1;
      cdc_hdr_dirty(cdc);
    }
  }
  return rc;
}

/*
 * NAME
 *   cdc_tbl_create
 *
 * DESCRIPTION
 *   create a new table
 *
 * PARAMETERS
 *   cdc : 
 *   tbl : table pointer
 *
 * RETURN
 *
 * NOTES
 */
int cdc_tbl_create(cdc_t *cdc, cdc_tbl_t *tbl)
{
  int           rc;
  unsigned long tbl_block;

  rc = 0;
  tbl_block
    = 1 
    + tbl->block_offset
    + tbl->count * cdc->map_distance;
  if (tbl_block != tbl->buffer.block)
  {
    rc = cdc_tbl_flush(cdc, tbl);
  }
  if (0 == rc)
  {
    int ii;

    tbl->buffer.block = tbl_block;
    tbl->buffer.dirty = 1;
    tbl->is_phantom    = 0;
    cdc_hdr_dirty(cdc);

    for (ii = 0; ii < tbl->entries; ii++)
    {
      tbl->buffer.data[ii] = tbl->phantom_value;
    }
    tbl->count++;
    /*rc = cdc_tbl_flush(cdc, tbl);*/
  }
  return rc;
}


