/*******************************************************************
 * cl-alt.cpp: Alternate version of CL program using system() call
 *			   instead of Win32 interprocess communication features
 *
 * Options: /silent
 *       or -silent:  disables STL filtering status message and auto-
 *                    matically adds /nologo option to CL2 command line
 *          /iter:n	  Sets iterator policy to n (S, M or L)
 *
 *
 * Calls cl2.exe to perform compilation, filters output to
 * simplify cryptic STL names (for files with applicable extensions,
 * as configured below)
 *
 * For installation instructions, see the Perl script (STLFilt.pl)
 *
 * Error filtering is toggled by the presence (or absence) of
 * the file named by STLFILE below.
 * Toggling the existence of this file is mananged either by the
 * STLFILT.BAT batch file or the STLTask taskbar icon app.
 *
 * Written by Leor Zolman (leor@bdsoft.com)
 *			January 2001
 *			Release 1.0 - 5/20/2001
 *
 * Note: Be sure not to compile this cl.cpp file using any version
 *       of CL.EXE resident in the current directory, or it won't
 *		 be able to overwrite itself to create the new CL.EXE...
 *******************************************************************/

#include <cstdio>					// use the C library to keep it small
#include <cstring>
#include <cstdlib>
#include <windows.h>

// Customize the following four strings, verbose option and the filterExt list.
// Note: If any pathnames contain embedded spaces, be sure to use the
// following form when defining them:
//		#define PATHNAME "\"c:\\program files\\whatever\""

#define PERL_EXE		"e:\\perl\\bin\\perl"		// Your Perl interpreter

#define FILTER_SCRIPT	"d:\\src\\cl\\STLFilt.pl"	// Perl script pathname

#define FILT_FILE		"d:\\src\\cl\\filtering.on"	// filtering toggle file in "on" mode
													//		(see companion variable in STLFILT.BAT)

#define STANDARD_CL		"cl2"						// name of standard CL.EXE, must
													//		reside in a system PATH dir

#define VERBOSE	1									// true to enable sign-on message
													//		(/silent option overrides)

const char *filterExt[] =							// Filter only output from compiling
{													// files with these extensions
	".cpp",
	".cxx",
	NULL
};


const char *objExt[] =								// If we see these extensions on any
{													// CL arguments at the end of the line,
	".lib",											// keep searching for some other
	".obj",											// extention to qualify for filtering...
	NULL
};

// Nothing to configure beyond this point.

#define MAX_ARGS (2 * MAX_PATH + 20)
#define CMD_LINE_TAIL	" | " PERL_EXE " " FILTER_SCRIPT

bool hasExt(const char *filename, const char *extensions[]);

int main(int argc, char **argv)
{
	FILE *fp;
	bool filtering = false;							// not filtering by default
	bool silent = false;							// not silent by default
	char *filename;									// file to compile
	char full_filt_tail[500] = CMD_LINE_TAIL;
	size_t i;
	
	if (argc == 1)
	{
		fprintf(stderr, "Proxy CL (Alt): Invoked with no command line args. Bailing out.\n");
		exit(1);
	}

	if (argv[1][0] == '@')							// indirect command line?
	{
		if ((fp = fopen(&argv[1][1], "r")) == NULL)	// yes, open args file
		{
			fprintf(stderr, "Proxy CL (Alt): Bad @ filename: %s; Bailing out.\n", argv[1]);
			exit(1);
		}

		char linebuf[500];

		while (fgets(linebuf, 500, fp) != NULL)
			;
		fclose(fp);

		bool seenquote = false;
		for (int i = strlen(linebuf) - 1; i >= 0; i--)
			if (linebuf[i] == '"')
				if (!seenquote)
				{
					seenquote = true;
					linebuf[i] = '\0';
				}
				else
					break;
			else if (linebuf[i] == ' ' && !seenquote)
				break;
		
		filename = &linebuf[i + 1];
	}
	else
		for (i = argc - 1; i >= 0; i--)
		{
			if (hasExt(argv[i], objExt ))			// ignore libraries, obj files
				continue;
			if (strchr(argv[i], '.') == NULL)		// ignore options (no extensions)
				continue;
			filename = argv[i];						// first non-lib/obj/option is the filename
			break;
		}

	
	if (!hasExt(filename, filterExt))				// if extension of filename
	{												// is not in the approved
		silent = true;								// list, never filter it
	}
	else if((fp = fopen(FILT_FILE, "r")) != NULL)
	{												// otherwise, filter iff
		filtering = true;							// FILT_FILE exists
		fclose(fp);
	}

	char args[MAX_ARGS] = " ";						// replicate CL command line args

	for (i = 1; i < argc; i++)
	{
		if (!stricmp(argv[i],"/silent") || !stricmp(argv[i],"-silent"))
		{
			silent = true;
			strcat(args, "/nologo");
		}
		else if (!strnicmp(argv[i], "/iter:", 6) || !strnicmp(argv[i], "-iter:", 6))
		{
			strcat(full_filt_tail, " ");
			strcat(full_filt_tail, argv[i]);		// pass iterator policy option to Perl
		}
		else
		{
			strcat(args,"\"");				// surround each arg with
			strcat(args, argv[i]);			// double quotes to avoid long
			strcat(args,"\"");				// filename problems
		}
		strcat(args, " ");
	}
										// Create replacement command line
	char cmdline[MAX_ARGS + 50] = STANDARD_CL;	// start with "old" CL.EXE
	strcat(cmdline, args);				// append whatever args were passed

  if (filtering)						
  {
#if VERBOSE								// announce the filtering
	if (!silent)
	{
		puts("     ****** STL Err0r Decryption (Alt) is ON! ******");
		fflush( stdout );
	}
#endif
	strcat(cmdline, full_filt_tail);	// pipe the errors to "Perl STLFilt.pl [/iter:n]"
	return system(cmdline);				// go do it
  }
  else
  {
#if VERBOSE								// announce the non-filtering
	if (!silent)
	{
		puts("     ****** STL Err0r Decryption (Alt) is OFF ******");
		fflush( stdout );
	}
#endif
	return system(cmdline);			// go do it
  }
}

//
// return true if the filename's extension is in extensions:
//

bool hasExt(const char *file, const char *extensions[])
{
	char filename[100];
	strcpy(filename, file);
	
	int i;
					// get rid of trailing whitespace on the filename...
	for (i = strlen(filename) - 1; i > 0; i--)
		if (isspace(filename[i]))
			filename[i] = '\0';
		else
			break;

	char buf[100];

	int length = strlen(filename);
	int extLen;
	for (i = 0; extensions[i] != 0; i++)
	{
		extLen = strlen(extensions[i]);
		if (length >= extLen + 1 &&
				!stricmp(&filename[strlen(filename) - extLen], extensions[i]))
			return true;
	}
	return false;
}
			
