/*******************************************************************
 * cl.cpp: "CL-Spoofing" STL compilation filter for VC++5/6
 *
 * Options: /silent
 *       or -silent:  disables STL filtering status message and auto-
 *                    matically adds /nologo option to CL2 command line
 *          /iter:x	  Sets iterator policy to n (S, M or L)
 *
 * Calls cl2.exe to perform compilation, filters output to
 * simplify cryptic STL names (for files with applicable extensions,
 * as configured below).
 *
 * For installation instructions, see the Perl script (STLFilt.pl)
 *
 * Error filtering is toggled by the presence (or absence) of
 * the file named by STLFILE below.
 * Toggling the existence of this file is mananged either by the
 * STLFILT.BAT batch file or the STLTask taskbar icon app.
 *
 * Framework written by Leor Zolman	(leor@bdsoft.com)
 *			Original version: January, 2001
 *			Current: Release 1.0 - 5/20/2001
 *
 * Low-level piping wizardry by Thomas Becker
 *
 * Note: Be sure not to compile this cl.cpp file using any version
 *       of CL.EXE resident in the current directory, or it won't
 *		 be able to overwrite itself to create the new CL.EXE...
 *******************************************************************/

#include <cstdio>					// use the C library to keep it small
#include <cstring>
#include <cstdlib>
#include <process.h>				// for _spawnvp()
#include <windows.h>				// needed to create processes

// Customize the following four strings, verbose option and the filterExt list.
// Note: If any pathnames contain embedded spaces, be sure to enclose
// such pathnames with literal double quotes, e.g.,
//		#define PATHNAME "\"c:\\program files\\whatever\""

#define PERL_EXE		"e:\\perl\\bin\\perl"		// Your Perl interpreter

#define FILTER_SCRIPT	"d:\\src\\cl\\STLFilt.pl"	// Perl script pathname

#define FILT_FILE		"d:\\src\\cl\\filtering.on"	// filtering toggle file in "on" mode
													//		(see companion variable in STLFILT.BAT)

#define STANDARD_CL		"cl2"						// name of standard CL.EXE, must
													//		reside in a system PATH dir

#define VERBOSE			1							// true to enable sign-on message
													//		(/silent option overrides)

const char *filterExt[] =							// Filter only output from compiling
{													// files with these extensions
	".cpp",
	".cxx",
	NULL
};


const char *objExt[] =								// If we see these extensions on any
{													// CL arguments at the end of the line,
	".lib",											// keep searching for some other
	".obj",											// extention to qualify for filtering...
	NULL
};

// Nothing to configure beyond this point.

#define MAX_ARGS	(2 * MAX_PATH + 20)
#define FILT_CMD	PERL_EXE " " FILTER_SCRIPT
#define DEBUG		0
#define DEBUG_LOG	"d:\\src\\cl\\debug.out"

BOOL WriteEofToPipe(HANDLE hPipe);
bool hasExt(const char *filename, const char *extensions[]);
void logit(const char *msg);

#if DEBUG
#define log(s) logit(s)
#else
#define log(s)
#endif

int main(int argc, char **argv)
{
	FILE *fp;
	bool filtering = false;							// not filtering by default
	bool silent = false;							// not silent by default
	char full_filt_cmd[500] = FILT_CMD;
	char *filename;									// file to compile
	size_t i;
	
	if (argc == 1)
	{
		fprintf(stderr, "Proxy CL: Invoked with no command line args. Bailing out.\n");
		exit(1);
	}
	
	if (argv[1][0] == '@')							// indirect command line?
	{
		log("Indirect command line detected:");
		log(argv[1]);

		if ((fp = fopen(&argv[1][1], "r")) == NULL)	// yes, open args file
		{
			log("Bad @ filename? First Command line arg:");
			log(argv[1]);
			fprintf(stderr, "Proxy CL: Bad @ filename: %s; Bailing out.\n", argv[1]);
			exit(1);
		}

		char linebuf[500];

		while (fgets(linebuf, 500, fp) != NULL)
			;
		fclose(fp);

		log("The complete line from the indirect file");
		log(linebuf);

		bool seenquote = false;
		for (int i = strlen(linebuf) - 1; i >= 0; i--)
			if (linebuf[i] == '"')
				if (!seenquote)
				{
					seenquote = true;
					linebuf[i] = '\0';
				}
				else
					break;
			else if (linebuf[i] == ' ' && !seenquote)
				break;
		
		filename = &linebuf[i + 1];

		log("filename detected indirectly:");
		log(filename);
	}
	else
		for (i = argc - 1; i >= 0; i--)
		{
			if (hasExt(argv[i], objExt ))			// ignore libraries, obj files
				continue;
			if (strchr(argv[i], '.') == NULL)		// ignore options (no extensions)
				continue;
			filename = argv[i];						// first non-lib/obj/option is the filename
			break;
		}
	
	if (!hasExt(filename, filterExt ))				// if extension of filename
	{												// is not in the approved
		silent = true;								// list, never filter it
	}
	else if((fp = fopen(FILT_FILE, "r")) != NULL)
	{												// otherwise, filter iff
		filtering = true;							// FILT_FILE exists
		fclose(fp);
	}

	char args[MAX_ARGS] = " ";						// replicate CL command line args
	for (i = 1; i < argc; i++)
	{
		if (!stricmp(argv[i],"/silent") || !stricmp(argv[i],"-silent"))
		{
			silent = true;
			strcat(args, "/nologo");
		}
		else if (!strnicmp(argv[i], "/iter:", 6) || !strnicmp(argv[i], "-iter:", 6))
		{
			strcat(full_filt_cmd, " ");
			strcat(full_filt_cmd, argv[i]);		// pass iterator policy option to Perl
		}
		else
		{
			strcat(args,"\"");				// surround each arg with
			strcat(args, argv[i]);			// double quotes to avoid long
			strcat(args,"\"");				// filename problems
		}

		strcat(args, " ");
	}

												// Create replacement command line
	char cmdline[MAX_ARGS + 50] = STANDARD_CL;	// start with "old" CL.EXE
	strcat(cmdline, args);						// append whatever args were passed
	log(cmdline);
	
  if (filtering)						
  {
#if VERBOSE								// announce the filtering
	if (!silent)
	{
		puts("     ****** STL Err0r Decryption is ON! ******");
		fflush( stdout );
	}
#endif
    
    // Get std device handles
    HANDLE hStdIn = GetStdHandle(STD_INPUT_HANDLE);
    HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
    HANDLE hStdError = GetStdHandle(STD_ERROR_HANDLE);
    
    // Create an anoymous pipe for communication between cl2.exe and perl.exe
    //
    HANDLE hPipeWrite = NULL;
    HANDLE hPipeRead = NULL;
    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof(sa); 
    sa.lpSecurityDescriptor = NULL; 
    sa.bInheritHandle = TRUE;
    if( ! ::CreatePipe(
      &hPipeRead,				// address of variable for read handle 
      &hPipeWrite,				// address of variable for write handle  
      &sa,						// pointer to security attributes 
      0							// number of bytes reserved for pipe 
      ) )
    {
      puts("CL.EXE: Failure to create pipe between cl2.exe and perl.exe");
      return 1;
    }
    
    // Start perl
    //
    PROCESS_INFORMATION pi_perl;
    pi_perl.hProcess = NULL;
    pi_perl.hThread = NULL;
    STARTUPINFO si_perl;
    memset(&si_perl, '\0', sizeof(si_perl));
    si_perl.cb = sizeof(si_perl);
    si_perl.dwFlags = STARTF_USESTDHANDLES;
    si_perl.hStdInput = hPipeRead;	// read side of pipe
    si_perl.hStdOutput = hStdOut;	// stdout of parent
    si_perl.hStdError = hStdOut;	// stderr of parent
    if( ! ::CreateProcess(
      NULL,
	  full_filt_cmd,				// something like: perl stlfilt.pl [/iter:n]
      NULL,							// pointer to process security attributes 
      NULL,							// pointer to thread security attributes 
      TRUE,							// handle inheritance flag 
      0,							// creation flags 
      NULL,							// pointer to new environment block 
      NULL,							// pointer to current directory name 
      &si_perl,						// pointer to STARTUPINFO 
      &pi_perl						// pointer to PROCESS_INFORMATION  
      ) )
    {
      puts("CL.EXE: Failure to start perl.exe (check pathname in CL.CPP)");
      return 1;
    }
    
    // Try-finally block for closing handles and sending EOF to perl.
    //
    PROCESS_INFORMATION pi_cl2 ;
    pi_cl2.hProcess = NULL;
    pi_cl2.hThread = NULL;
    //
    __try
    {
      // Start cl2.exe
      //
      STARTUPINFO si_cl2;
      memset(&si_cl2, '\0', sizeof(si_cl2));
      si_cl2.cb = sizeof(si_cl2);
      si_cl2.dwFlags = STARTF_USESTDHANDLES;
      si_cl2.hStdInput = hStdIn;		// stdin of parent
      si_cl2.hStdOutput = hPipeWrite;	// write side of child-to-parent
      si_cl2.hStdError = hStdError;		// stderr of parent
      if( ! ::CreateProcess(NULL,
        cmdline,
        NULL,							// pointer to process security attributes 
        NULL,							// pointer to thread security attributes 
        TRUE,							// handle inheritance flag 
        0,								// creation flags 
        NULL,							// pointer to new environment block 
        NULL,							// pointer to current directory name 
        &si_cl2,						// pointer to STARTUPINFO 
        &pi_cl2							// pointer to PROCESS_INFORMATION  
        ) )
      {
        puts("CL.EXE: Failure to start cl2.exe (does your PATH include it?)");
        return 1;
      }
      
      // Wait for cl2.exe to finish.
      if( WAIT_FAILED == ::WaitForSingleObject(pi_cl2.hProcess, INFINITE) )
      {
        puts("CL.EXE: Failure waiting for cl2.exe to terminate");
        return 1;
      }
    }
    __finally
    {
      
      // Send EOF to perl to make it end processing and exit.
      if( ! WriteEofToPipe(hPipeWrite) )
      {
        puts("CL.EXE: Failure to send EOF to perl.exe");
        return 1;
      }
    }
    
    // Wait for perl.exe to finish.
    if( WAIT_FAILED == ::WaitForSingleObject(pi_perl.hProcess, INFINITE) )
    {
      puts("CL.EXE: Failure waiting for perl.exe to terminate");
      return 1;
    }
    
    // Get exit code for cl2.exe. This is possible because we have not closed
    // the process handle yet.
    //
    DWORD dwExitCode;
    if( ! GetExitCodeProcess(
      pi_cl2.hProcess,		// handle to the process
      &dwExitCode			// address to receive termination status
      ) )
    {
      puts("CL.EXE: Could not get exit code for cl2.exe");
      return 1;
    }
    
    // Close all handles, just to make the point.
    //
    ::CloseHandle(hPipeWrite);
    ::CloseHandle(hPipeRead);
    ::CloseHandle(pi_perl.hProcess);
    ::CloseHandle(pi_perl.hThread);
    ::CloseHandle(pi_cl2.hProcess);
    ::CloseHandle(pi_cl2.hThread);
    
    return dwExitCode;
  }
  else
  {
#if VERBOSE					// announce the non-filtering
	if (!silent)
	{
		puts("     ****** STL Err0r Decryption is OFF ******");
		fflush( stdout );
	}
#endif
							// if not filtering, use old CL:
	int ret_stat = _spawnvp(_P_WAIT, STANDARD_CL, argv);
	if (ret_stat == -1)
	{
		puts("CL.EXE: Failure to start cl2.exe (does your PATH include it?)");
		return 1;
	}
	else
		return ret_stat;
  }
}

BOOL WriteEofToPipe(HANDLE hPipe)
{
		DWORD dwNumBytesWritten = 0 ;
		BOOL bWriteRetVal = WriteFile(
		  hPipe,				// handle to file to write to 
		  "\032",				// pointer to data to write to file 
		  1,					// number of bytes to write 
		  &dwNumBytesWritten,	// pointer to number of bytes written 
		  NULL) ;				// addr. of structure needed for overlapped I/O  
		if( ! bWriteRetVal || 1 != dwNumBytesWritten )
		  return FALSE;

    ::FlushFileBuffers(hPipe);
    return TRUE;
}

//
// return true if the filename's extension is in extensions:
//

bool hasExt(const char *file, const char *extensions[])
{
	char filename[100];
	strcpy(filename, file);
	
	int i;
					// get rid of trailing whitespace on the filename...
	for (i = strlen(filename) - 1; i > 0; i--)
		if (isspace(filename[i]))
			filename[i] = '\0';
		else
			break;

	char buf[100];
	sprintf(buf, "filename being examined: \"%s\"", filename);
	log(buf);

	int length = strlen(filename);
	int extLen;
	for (i = 0; extensions[i] != 0; i++)
	{
		extLen = strlen(extensions[i]);
		if (length >= extLen + 1 &&
				!stricmp(&filename[strlen(filename) - extLen], extensions[i]))
		{
			log("okToFilter true for:");
			log(filename);
			return true;
		}
	}
	log("okToFilter false for:");
	log(filename);
	return false;
}
			
void logit(const char *msg)
{
	FILE *fp;
	
	if ((fp = fopen(DEBUG_LOG, "a")) == NULL)
		fprintf(stderr, "Can't append to debug.out.");
	else {
		fputs(msg, fp);
		fputs("\n", fp);
		fclose(fp);
	}
}
	
