// file: base.cpp
// Implementation of the Base class.

#ifdef _MSC_VER
// Disable MS warnings about long identifiers
#pragma warning(disable:4786)
#endif

#include "base.h"
#include <iostream>
#include <strstream>
#include "factory.h"

// Helper to keep maintenance down.  This is the unique
// identifier for the class.  It could just as easily be a class static
// variable.
const Base::BASE_KEY_TYPE baseName ("Base");

// The unique identifier for the class.  Now, you may be wondering
// why not use rtti, as in: 'typeid(*this).name();'  The reason is,
// typeid().name() returns a compiler-specific string.  That would
// tie this class, and its streamed type, to a specific compiler,
// or potentially worse a specific version of a specific compiler.
// By defining our own type, we are now platform-independent.  The
// down-side, of course, is that now we have to make sure each type
// has a unique identifier - hence I'm using the class name.
Base::BASE_KEY_TYPE Base::classID() const
{
    return baseName;
}

// Register the creation function with the factory.
namespace
{
    // Note that this highlights one of the dangers of this technique.  We have
    // no control over the order in which various modules will call regCreateFn,
    // so this technique cannot be safely used if the factory is required
    // to instantiate an object before main() executes.
    registerInFactory<Base, Base> registerMe(baseName);
}

// Base and its subclasses use the two-stage initialization
// process.  init() forwards the work to initialize().  Each
// child class must call its parent's initialize() function,
// therefore initialize() is protected.
void Base::init()
{
    initialize();
}

// Implementation specifics - get a number from the user.
void Base::initialize()
{
    // OK, it's hokey, whaddya want for a simple example? :-)
    std::cout << "Enter a number for Base: ";
    std::cin >> myValue;
}

// read() just follows the Template Method, and forwards the call to
// the virtual do_read() function.  This is overkill for this example,
// but in a "real" implementation this would perform some basic
// error checking to ensure that the file is valid, before passing
// control over to do_read().
//
// Child classes must call their parents' do_read() function.
void Base::read(std::istream &i)
{
    do_read(i);
}

// Similar to read(), write() follows the Template Method.  In this
// case, we actually do some preprocessing before writing out
// the class itself.  We write the most-derived class's type to
// the stream, then stream ourself out.  Then, when the factory
// loads it in, the factory picks out the type from the stream
// and then creates the appropriate type from that.
void Base::write(std::ostream &o) const
{
    o << classID() << std::endl;
    do_write(o);
}


////////////////////////////////////////////////////////////////////////
// Virtual function implementations
void Base::do_write(std::ostream &o) const
{
    o << "\t" << myValue << std::endl;
}

void Base::do_read(std::istream &instr)
{
    instr >> myValue;

    // Skip rest of line
    std::string inputLine;
    std::getline(instr, inputLine);
}

void Base::displayValues()
{
    std::cout << "Hi, I'm a Base class and my value is " << myValue << "." << std::endl;
}

Base::~Base()
{
}

////////////////////////////////////////////////////////////////////////
// Simple implementation of the insertion/extraction operators.
//
std::ostream & operator <<(std::ostream &o, const Base &b)
{
    b.write(o);
    return o;
}

std::istream & operator >>(std::istream &o, Base &b)
{
    b.read(o);
    return o;
}

