// main.cpp
// Copyright 2001 Jim Hyslop, all rights reserved.
// Example and test file to illustrate how to use the factory template.

#ifdef _MSC_VER
// Disable MS warnings about long identifiers
#pragma warning(disable:4786)
#endif

#include <iostream>
#include <fstream>
#include "base.h"
#include <vector>
#include "factory.h"

// In a production environment, there would of course be much stricter
// error checking.
void loadInputFile(const std::string &inputFileName,
                   std::vector<Base *> &loadedClasses)
{
    int count;

    std::ifstream infile(inputFileName.c_str());
    for (count=0; count < loadedClasses.size(); ++count) {
        delete loadedClasses[count];
    }
    loadedClasses.clear();
    int numClasses;

    // File format is very simple. It is a text file, consisting of a file
    // header, followed by the information required to restore each object
    // properly. It's not the most efficient, but if you open the output
    // file in a text editor it becomes quite clear.
    // File header consists of the number of objects stored on a single line.
    // Object information consists of the object identifier token (in this case,
    // the object name) on a line by itself, followed by the actual data.
    infile >> numClasses;
    loadedClasses.resize(numClasses);

    // Ignore the rest of the line
    std::string tmp;
    std::getline(infile, tmp);

    // Reload the array now.
    for (count=0; count < loadedClasses.size(); count++) {
        // Step 1: Retrieve the identifier from the file.
        std::string className;
        std::getline(infile, className);

        // Step 2: Pass that identifier to the factory, to get
        // a default-initialized instance of that class.
        std::auto_ptr<Base> newBase(genericFactory<Base>::instance().create(className));
        loadedClasses[count] = newBase.release();

        // Step 3: If the factory recognized the identifier,
        // restore the state of the object.
        if (loadedClasses[count] != 0) {
            infile >> *loadedClasses[count];
        }
    }
}

void saveOutputFile(const std::string &saveFileName, 
                    std::vector<Base *> &loadedClasses)
{
    std::ofstream outfile(saveFileName.c_str());
    outfile << loadedClasses.size() << " objects in the list" << std::endl;
    int count;
    
    for (count=0; count < loadedClasses.size(); count++) {
        outfile << *(loadedClasses[count]);
    }
}

Base * createNewObject(const Base::BASE_KEY_TYPE &objectType)
{
    std::auto_ptr<Base> newBase = genericFactory<Base>::instance().create(objectType);
    if (newBase.get()) {
        newBase->init();
    }
    return newBase.release();
}


const int extraMenuItemCount=4;
int main()
{
    int menuCount;
    int userChoice;
    const std::string saveFileName("save.txt");

    // Get a list of the classes that are registered, to add to the menu options
    std::vector<std::string> classes (genericFactory<Base>::instance().getRegisteredClasses());

    // I'd use auto_ptr instead of raw pointers, but auto_ptrs can't be put
    // inside a standard container.
    std::vector<Base *> loadedClasses;
    do {
        int count;
        // Note the lack of error checking - to crash this program, enter a number
        // outside the expected range.
        std::cout << "\nEnter a choice\n"
            "0: quit\n"
            "1: Load file\n"
            "2: Save file\n"
            "3: Call virtual functions on all objects\n";
        for (menuCount=0; menuCount<classes.size(); menuCount++) {
            std::cout << menuCount+extraMenuItemCount << 
                ": Create new object of type " << classes[menuCount] << std::endl;
        }
        menuCount+=extraMenuItemCount;
        std::cin >> userChoice;
        switch (userChoice) {
        case 1:
            loadInputFile(saveFileName, loadedClasses);
            break;
        case 2:
            saveOutputFile(saveFileName, loadedClasses);
            break;
        case 3:
            for (count=0; count < loadedClasses.size(); ++count) {
                loadedClasses[count]->displayValues();
            }
            break;
        case 0:
            // do nothing
            break;
        default:
            Base * newBase;
            newBase = createNewObject(classes[userChoice-extraMenuItemCount]);
            loadedClasses.push_back(newBase);
            break;
        }
    } while (userChoice != 0);

    for (int count=0; count < loadedClasses.size(); ++count) {
        delete loadedClasses[count];
    }
    return 0;
}

